/* word.c   by Michael Thorpe   2017-04-12 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lisp.h"

static obj *already_defined_error,*undefined_error;

#define HASH_SIZE 127

static struct wordextra *hashtable[HASH_SIZE];

static unsigned long hash(const char *data,unsigned long len) {
	unsigned long h=0,m=1;

	while(len--) {
		h=(h+m*data++[0])%HASH_SIZE;
		m=(m*3)%HASH_SIZE;
	}
	return(h);
}

static void drop_word(obj *o) {
	struct wordextra *entry;
	unsigned long h;

	entry=o->value.word.extra;
	if(entry->previnhash)
		entry->previnhash->nextinhash=entry->nextinhash;
	else {
		h=hash(entry->data,entry->len);
		assert(hashtable[h]==entry);
		hashtable[h]=entry->nextinhash;
	}
	if(entry->nextinhash)
		entry->nextinhash->previnhash=entry->previnhash;
	free(entry->data);
	free(entry);
/* o->value must be 0 here, as a definition INCREF's its word */
}

static obj *word_tostring(obj *in) {
	obj *out;

	out=new_string(in->value.word.extra->data,in->value.word.extra->len);
	DECREF(in);
	return(out);
}

obj *define(obj *name,obj *val) {
	if(&word_objtype != name->type) {
		DECREF(val);
		return(throwtypeerror(name));
	}
	if(name->value.word.value) {
		DECREF(val);
		return(throw(INCREF(already_defined_error),name));
	}
	name->value.word.value=val;
	return(INCREF(name)); /* Having a defined value counts as a reference */
}

obj *lookup(obj *name) {
	obj *val;

	if(&word_objtype != name->type)
		return(throwtypeerror(name));
	if(!name->value.word.value)
		return(throw(INCREF(undefined_error),name));
	val=INCREF(name->value.word.value);
	DECREF(name);
	return(val);
}

obj *set(obj *name,obj *val) {
	if(&word_objtype != name->type) {
		DECREF(val);
		return(throwtypeerror(name));
	}
	if(!name->value.word.value) {
		DECREF(val);
		return(throw(INCREF(undefined_error),name));
	}
	DECREF(name->value.word.value);
	name->value.word.value=val;
	return(name);
}

obj *undefine(obj *name) {
	obj *val;

	if(&word_objtype != name->type)
		return(throwtypeerror(name));
	if(!name->value.word.value)
		return(throw(INCREF(undefined_error),name));
	val=name->value.word.value;
	name->value.word.value=0;
	DECREF(name); /* Having a defined value counts as a reference */
	DECREF(name); /* And we're throwing away a "real" reference */
	return(val);
}

obj *new_word(const char *data,unsigned long len) {
	obj *o;
	char *s;
	struct wordextra *entry;
	unsigned long h;

	h=hash(data,len);
	for(entry=hashtable[h];entry;entry=entry->nextinhash)
		if(entry->len==len && !memcmp(entry->data,data,len))
			break;
	if(entry)
		return(INCREF(entry->obj));
	s=(char *)malloc(len);
	if(!s)
		return(throwoom());
	entry=(struct wordextra *)malloc(sizeof(struct wordextra));
	if(!entry) {
		free(s);
		return(throwoom());
	}
	o=new_object(&word_objtype);
	if(!o) {
		free(entry);
		free(s);
		return(0);
	}
	memcpy(s,data,len);
	o->value.word.extra=entry;
	o->value.word.value=0;
	entry->data=s;
	entry->len=len;
	entry->obj=o;
	entry->previnhash=0;
	entry->nextinhash=hashtable[h];
	if(hashtable[h])
		hashtable[h]->previnhash=entry;
	hashtable[h]=entry;
	return(o);
}

int store_builtin(unsigned int numargs,const char *name,obj *(*rawfunc)()) {
	obj *key,*func,*o;

	key=new_word(name,strlen(name));
	if(!key)
		return(-1);
	func=new_object(&rawfunc_objtype);
	if(!func) {
		DECREF(key);
		return(-1);
	}
	func->value.rawfunc.numargs=numargs;
	func->value.rawfunc.rawfuncptr.ns0=rawfunc;
	o=define(key,func);
	DECREF(o);
	return(!o);
}

int store_object(const char *name,obj *o) {
	obj *key;

	key=new_word(name,strlen(name));
	if(!key)
		return(-1);
	o=define(key,o);
	if(!o)
		return(-1);
	DECREF(o);
	return(0);
}

obj *unique_word() {
	int len;
	unsigned long h;
	struct wordextra *entry;
	char buf[1+4+1+10+1];

	while(1) {
		len=snprintf(buf,sizeof(buf),"#word-%u",rand());
		assert(6<len);
		assert(len<sizeof(buf));
		h=hash(buf,len);
		for(entry=hashtable[h];entry;entry=entry->nextinhash)
			if(entry->len==len && !memcmp(entry->data,buf,len))
				break;
		if(!entry)
			return(new_word(buf,len));
	}
}

static int word_init() {
	obj *f;

	if(store_builtin_ns0("unique-word",unique_word)
	|| !(already_defined_error=new_word("already-defined",15))
	|| !(f=new_word("#f",2))
	|| !(undefined_error=new_word("undefined",9)))
		return(1);
	DECREF(define(f,INCREF(&nilobj)));
	return(0);
}

static void word_exit() {
	struct wordextra *entry;
	unsigned long h;

	for(h=0;h<HASH_SIZE;h++) {
		for(entry=hashtable[h];entry;entry=entry->nextinhash) {
			if(entry->obj->value.word.value) {
				DECREF(entry->obj->value.word.value);
				entry->obj->value.word.value=0;
				DECREF(entry->obj);
			}
		}
	}
	DECREF(already_defined_error); already_defined_error=0;
	DECREF(undefined_error); undefined_error=0;
}

objtype word_objtype={
	name: "word",
	tostring: word_tostring,
	predropobj: drop_word,
	init: word_init,
	exit: word_exit,
};

